from loguru import logger
from rich.highlighter import ReprHighlighter
from rich.pretty import Pretty, is_expandable

from flexget import options, plugin
from flexget.event import event
from flexget.terminal import TerminalTable, console

logger = logger.bind(name='dump')


def dump(entries, debug=False, eval_lazy=False, trace=False, title_only=False):
    """
    Dump *entries* to stdout

    :param list entries: Entries to be dumped.
    :param bool debug: Print non printable fields as well.
    :param bool eval_lazy: Evaluate lazy fields.
    :param bool trace: Display trace information.
    :param bool title_only: Display only title field
    """

    def sort_key(field):
        # Sort certain fields above the rest
        if field == 'title':
            return (0,)
        if field == 'url':
            return (1,)
        if field == 'original_url':
            return (2,)
        return 3, field

    highlighter = ReprHighlighter()

    for entry in entries:
        entry_table = TerminalTable(
            'field',
            ':',
            'value',
            show_header=False,
            show_edge=False,
            pad_edge=False,
            collapse_padding=True,
            box=None,
            padding=0,
        )
        for field in sorted(entry, key=sort_key):
            if field.startswith('_') and not debug:
                continue
            if title_only and field != 'title':
                continue
            if entry.is_lazy(field) and not eval_lazy:
                renderable = (
                    '[italic]<LazyField - value will be determined when it is accessed>[/italic]'
                )
            else:
                try:
                    value = entry[field]
                except KeyError:
                    renderable = '[italic]<LazyField - lazy lookup failed>[/italic]'
                else:
                    if field.rsplit('_', maxsplit=1)[-1] == 'url':
                        renderable = f'[link={value}][repr.url]{value}[/repr.url][/link]'
                    elif isinstance(value, str):
                        renderable = value.replace('\r', '').replace('\n', '')
                    elif is_expandable(value):
                        renderable = Pretty(value)
                    else:
                        try:
                            renderable = highlighter(str(value))
                        except Exception:
                            renderable = f'[[i]not printable[/i]] ({repr(value)})'
            entry_table.add_row(f'{field}', ': ', renderable)
        console(entry_table)
        if trace:
            console('── Processing trace:', style='italic')
            trace_table = TerminalTable(
                'Plugin',
                'Operation',
                'Message',
                show_edge=False,
                pad_edge=False,
            )
            for item in entry.traces:
                trace_table.add_row(item[0], '' if item[1] is None else item[1], item[2])
            console(trace_table)
        if not title_only:
            console('')


class OutputDump:
    """
    Outputs all entries to console
    """

    schema = {'type': 'boolean'}

    @plugin.priority(0)
    def on_task_output(self, task, config):
        if not config and task.options.dump_entries is None:
            return

        eval_lazy = 'eval' in task.options.dump_entries
        trace = 'trace' in task.options.dump_entries
        title = 'title' in task.options.dump_entries
        states = {'accepted': 'green', 'rejected': 'red', 'failed': 'yellow', 'undecided': 'none'}
        dumpstates = [s for s in states if getattr(task, s)]
        specificstates = [s for s in states if s in task.options.dump_entries]
        if specificstates:
            dumpstates = specificstates
        for state in dumpstates:
            console.rule(state.title(), style=states[state])
            if getattr(task, state):
                dump(getattr(task, state), task.options.debug, eval_lazy, trace, title)
            else:
                console(f'No {state} entries', style='italic')
        if not dumpstates:
            console('No entries were produced', style='italic')


@event('plugin.register')
def register_plugin():
    plugin.register(OutputDump, 'dump', builtin=True, api_ver=2)


@event('options.register')
def register_parser_arguments():
    options.get_parser('execute').add_argument(
        '--dump',
        nargs='*',
        choices=['eval', 'trace', 'accepted', 'rejected', 'undecided', 'failed', 'title'],
        dest='dump_entries',
        help=(
            'display all entries in task with fields they contain, '
            'use `--dump eval` to evaluate all lazy fields. Specify an entry '
            'state/states to only dump matching entries.'
        ),
    )



import random
import re
import time
from urllib.parse import quote

from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.cached_input import cached
from flexget.utils.simple_persistence import SimplePersistence
from flexget.utils.soup import get_soup
from flexget.utils.tools import parse_filesize

logger = logger.bind(name='magnetdl')


class Page404Error(Exception):
    pass


class MagnetDL:
    """Creates entries from magnetdl categories"""

    url = 'https://www.magnetdl.com'

    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {
                'type': 'object',
                'properties': {
                    'category': {
                        'type': 'string',
                        'enum': ['software', 'movies', 'games', 'e-books', 'tv', 'music'],
                    },
                    'pages': {'type': 'integer', 'minimum': 1, 'maximum': 30, 'default': 5},
                },
                'additionalProperties': False,
            },
        ]
    }

    def _url(self, category, page):
        return self.url + '/download/' + category + '/' + (str(page) if page > 0 else '')

    def parse_page(self, scraper, url: str):
        try:
            logger.debug('page url: {}', url)
            page = scraper.get(url)
        except RequestException as e:
            raise plugin.PluginError(str(e))
        if page.status_code == 404:
            raise Page404Error()
        if page.status_code != 200:
            raise plugin.PluginError(f'HTTP Request failed {page.status_code}. Url: {url}')

        soup = get_soup(page.text)
        soup_table = soup.find('table', class_='download')
        if not soup_table:
            # very likely no result
            return
        table_tbody = soup_table.find('tbody')
        if not table_tbody:
            raise plugin.PluginError('Parsing crashed, no tbody, please report the issue')

        trs = table_tbody.find_all('tr')
        if not trs:
            logger.critical('Nothing to parse')
            return
        for tr in trs:
            try:
                magnet_td = tr.find('td', class_='m')
                if not magnet_td:
                    # skip empty trs
                    continue
                magnet_a = magnet_td.find('a')
                magnet = magnet_a['href']
                title_td = tr.find('td', class_='n')
                title_a = title_td.find('a')
                title = title_a['title']
                seed_td = tr.find('td', class_='s')
                seed = int(seed_td.text)
                leech = int(tr.find('td', class_='l').text)
                content_size = parse_filesize(seed_td.previous_sibling.text)
                yield Entry(
                    url=magnet,
                    title=title,
                    torrent_seeds=seed,
                    torrent_leech=leech,
                    content_size=content_size,
                )
            except AttributeError as e:
                raise plugin.PluginError('Parsing crashed, please report the issue') from e

    @cached('magnetdl', persist='4 minutes')
    def on_task_input(self, task, config):
        try:
            import cloudscraper
        except ImportError as e:
            logger.debug('Error importing cloudscraper: {}', e)
            raise plugin.DependencyError(
                issued_by='cfscraper',
                missing='cloudscraper',
                message='CLOudscraper module required. ImportError: %s' % e,
            )

        scraper = cloudscraper.create_scraper()
        category = config['category']
        persistence = SimplePersistence(plugin='magnetdl')
        last_magnet = persistence.get(category, None)
        logger.debug('last_magnet: {}', last_magnet)
        first_magnet = None
        stop = False

        for page in range(0, config['pages']):
            logger.verbose('Retrieving {} page {}', category, page + 1)
            url = self._url(category, page)
            logger.debug('Url: {}', url)
            try:
                for entry in self.parse_page(scraper, url):
                    if first_magnet is None:
                        first_magnet = entry['url']
                        logger.debug('Set first_magnet to {}', first_magnet)
                        persistence[category] = first_magnet
                    if last_magnet == entry['url']:
                        logger.debug('Found page where we have left, stopping')
                        stop = True
                    yield entry
            except Page404Error:
                logger.warning('Page {} returned 404, stopping', page)
                return
            if stop:
                return
            time.sleep(random.randint(1, 5))

    # Search API method
    def search(self, task, entry, config):
        if not config:
            return
        try:
            import cloudscraper
        except ImportError as e:
            logger.debug('Error importing cloudscraper: {}', e)
            raise plugin.DependencyError(
                issued_by='cfscraper',
                missing='cloudscraper',
                message='CLOudscraper module required. ImportError: %s' % e,
            )
        scraper = cloudscraper.create_scraper()
        entries = []
        for search_string in entry.get('search_strings', [entry['title']]):
            logger.debug('Searching `{}`', search_string)
            try:
                term = search_string.lower().replace(' ', '-')
                # note: weird url convention, uses first letter of search term
                slash = term[0]
                url = 'https://www.magnetdl.com/{}/{}/'.format(slash, term)
                for entry in self.parse_page(scraper, url):
                    entries.append(entry)
            except Page404Error:
                logger.warning('Url {} returned 404', url)
                return entries
            time.sleep(random.randint(1, 5))
        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(MagnetDL, 'magnetdl', api_ver=2, interfaces=['task', 'search'])



from flexget import options, plugin
from flexget.event import event
from flexget.manager import Manager
from flexget.terminal import TerminalTable, console, table_parser
from flexget.utils.database import with_session

from . import db

try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.imdb.utils import extract_id, is_imdb_url
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='imdb')


def do_cli(manager, options):
    if options.seen_action == 'forget':
        seen_forget(manager, options)
    elif options.seen_action == 'add':
        seen_add(manager, options)
    elif options.seen_action == 'search':
        seen_search(manager, options)


def seen_forget(manager: Manager, options):
    forget_name = options.forget_value
    if is_imdb_url(forget_name):
        imdb_id = extract_id(forget_name)
        if imdb_id:
            forget_name = imdb_id

    tasks = None
    if options.tasks:
        tasks = []
        for task in options.tasks:
            try:
                tasks.extend(m for m in manager.matching_tasks(task) if m not in tasks)
            except ValueError as e:
                console(e)
                continue

    # If tasks are specified it should use pattern matching as search
    if tasks:
        forget_name = forget_name.replace("%", "\\%").replace("_", "\\_")
        forget_name = forget_name.replace("*", "%").replace("?", "_")

    count, fcount = db.forget(forget_name, tasks=tasks, test=options.test)
    console(f'Removed {count} titles ({fcount} fields)')
    manager.config_changed()


def seen_add(manager: Manager, options):
    DEFAULT_TASK = 'cli_add'

    seen_name = options.add_value
    if is_imdb_url(seen_name):
        console('IMDB url detected, try to parse ID')
        imdb_id = extract_id(seen_name)
        if imdb_id:
            seen_name = imdb_id
        else:
            console("Could not parse IMDB ID")

    task = DEFAULT_TASK
    local = None
    if options.task and not options.task in manager.tasks:
        console(f"Task `{options.task}` not in config")
        return
    else:
        task = options.task
        local = True

    db.add(seen_name, task, {'cli_add': seen_name}, local=local)

    if task == DEFAULT_TASK:
        console(f'Added `{seen_name}` as seen. This will affect all tasks.')
    else:
        console(f'Added `{seen_name}` as seen. This will affect `{task}` task.')


@with_session
def seen_search(manager: Manager, options, session=None):
    search_term = options.search_term
    if is_imdb_url(search_term):
        console('IMDB url detected, parsing ID')
        imdb_id = extract_id(search_term)
        if imdb_id:
            search_term = imdb_id
        else:
            console("Could not parse IMDB ID")
    else:
        search_term = search_term.replace("%", "\\%").replace("_", "\\_")
        search_term = search_term.replace("*", "%").replace("?", "_")

    tasks = None
    if options.tasks:
        tasks = []
        for task in options.tasks:
            try:
                tasks.extend(m for m in manager.matching_tasks(task) if m not in tasks)
            except ValueError as e:
                console(e)
                continue

    seen_entries = db.search(value=search_term, status=None, tasks=tasks, session=session)
    table = TerminalTable('Field', 'Value', table_type=options.table_type)
    for se in seen_entries.all():
        table.add_row('Title', se.title)
        for sf in se.fields:
            if sf.field.lower() == 'title':
                continue
            table.add_row('{}'.format(sf.field.upper()), str(sf.value))
        table.add_row('Task', se.task)
        if se.local:
            table.add_row('Local', 'Yes')
        table.add_row('Added', se.added.strftime('%Y-%m-%d %H:%M'), end_section=True)
    if not table.rows:
        console('No results found for search')
        return
    console(table)


@event('options.register')
def register_parser_arguments():
    parser = options.register_command(
        'seen', do_cli, help='View or forget entries remembered by the seen plugin'
    )
    subparsers = parser.add_subparsers(dest='seen_action', metavar='<action>')
    forget_parser = subparsers.add_parser(
        'forget', help='Forget entry or entire task from seen plugin database'
    )
    forget_parser.add_argument(
        '--tasks',
        nargs='+',
        metavar='TASK',
        help='forget only in specified task(s), optionally using glob patterns ("tv-*"). '
        'matching is case-insensitive',
    )

    forget_parser.add_argument(
        'forget_value',
        metavar='<value>',
        help='Title or url of entry to forget, or name of task to forget',
    )
    add_parser = subparsers.add_parser('add', help='Add a title or url to the seen database')
    add_parser.add_argument(
        '--task',
        metavar='TASK',
        help='add in specified task'
        'matching is case-insensitive. Will make entry local to that task',
    )
    add_parser.add_argument('add_value', metavar='<value>', help='the title or url to add')
    search_parser = subparsers.add_parser(
        'search', help='Search text from the seen database', parents=[table_parser]
    )
    search_parser.add_argument(
        '--tasks',
        nargs='+',
        metavar='TASK',
        help='search only in specified task(s), optionally using glob patterns ("tv-*"). '
        'matching is case-insensitive',
    )
    search_parser.add_argument('search_term', metavar='<search term>')



from datetime import datetime

from dateutil.parser import ParserError, isoparse
from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import RequestException, Session, TimedLimiter

plugin_name = 'discord'

logger = logger.bind(name=plugin_name)
session = Session()
session.add_domain_limiter(TimedLimiter('discord.com', '3 seconds'))


class DiscordNotifier:
    """
    Example::

      notify:
        entries:
          via:
            - discord:
                web_hook_url: <string>
                [username: <string>] (override the default username of the webhook)
                [avatar_url: <string>] (override the default avatar of the webhook)
                [embeds: <arrays>[<object>]] (override embeds)
    """

    schema = {
        'type': 'object',
        'properties': {
            'web_hook_url': {'type': 'string', 'format': 'uri'},
            'username': {'type': 'string', 'default': 'Flexget'},
            'avatar_url': {'type': 'string', 'format': 'uri'},
            'embeds': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'title': {'type': 'string'},
                        'description': {'type': 'string'},
                        'url': {'type': 'string', 'format': 'uri'},
                        'color': {'type': 'integer'},
                        'footer': {
                            'type': 'object',
                            'properties': {
                                'text': {'type': 'string'},
                                'icon_url': {'type': 'string', 'format': 'uri'},
                                'proxy_icon_url': {'type': 'string', 'format': 'uri'},
                            },
                            'required': ['text'],
                            'additionalProperties': False,
                        },
                        'image': {
                            'type': 'object',
                            'properties': {
                                'url': {'type': 'string', 'format': 'uri'},
                                'proxy_url': {'type': 'string', 'format': 'uri'},
                            },
                            'additionalProperties': False,
                        },
                        'thumbnail': {
                            'type': 'object',
                            'properties': {
                                'url': {'type': 'string', 'format': 'uri'},
                                'proxy_url': {'type': 'string', 'format': 'uri'},
                            },
                            'additionalProperties': False,
                        },
                        'timestamp': {'type': 'string'},
                        'provider': {
                            'type': 'object',
                            'properties': {
                                'name': {'type': 'string'},
                                'url': {'type': 'string', 'format': 'uri'},
                            },
                            'additionalProperties': False,
                        },
                        'author': {
                            'type': 'object',
                            'properties': {
                                'name': {'type': 'string'},
                                'url': {'type': 'string', 'format': 'uri'},
                                'icon_url': {'type': 'string', 'format': 'uri'},
                                'proxy_icon_url': {'type': 'string', 'format': 'uri'},
                            },
                            'additionalProperties': False,
                        },
                        'fields': {
                            'type': 'array',
                            'minItems': 1,
                            'items': {
                                'type': 'object',
                                'properties': {
                                    'name': {'type': 'string'},
                                    'value': {'type': 'string'},
                                    'inline': {'type': 'boolean'},
                                },
                                'required': ['name', 'value'],
                                'additionalProperties': False,
                            },
                        },
                    },
                    'additionalProperties': False,
                },
            },
        },
        'required': ['web_hook_url'],
        'additionalProperties': False,
    }

    def notify(self, title, message, config):
        """
        Send discord notification

        :param str message: message body
        :param dict config: discord plugin config
        """

        for embed in config.get('embeds', []):
            ts = embed.get('timestamp')
            if ts:
                if isinstance(ts, str):
                    if ts.isdigit():
                        try:
                            ts = datetime.utcfromtimestamp(int(ts))
                        except (ValueError, OverflowError):
                            logger.info(
                                f"Value provided for 'timestamp' ({embed['timestamp']}) "
                                f"is not a timestamp ({int(datetime.now().timestamp())})."
                            )
                    else:
                        try:
                            ts = isoparse(ts)
                            embed['timestamp'] = ts
                        except (ParserError, ValueError) as e:
                            logger.info(f"'timestamp' is in an invalid format: {e}")
                if not isinstance(ts, datetime):
                    embed.pop('timestamp', None)
                    logger.warning("'timestamp' is invalid, dropping it")
                else:
                    embed['timestamp'] = datetime.strftime(ts, r'%Y-%m-%dT%H:%M:%S%z')

        web_hook = {
            'content': message,
            'username': config.get('username'),
            'avatar_url': config.get('avatar_url'),
            'embeds': config.get('embeds'),
        }

        try:
            session.post(config['web_hook_url'], json=web_hook)
        except RequestException as e:
            raise PluginWarning(e.args[0])


@event('plugin.register')
def register_plugin():
    plugin.register(DiscordNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession

requests = RequestSession(max_retries=3)

plugin_name = 'bark'

logger = logger.bind(name=plugin_name)


class BarkNotifier:
    """
    Example::

      notify:
        entries:
        title: |
              {{task}}
        message: |
              {% if series_name is defined %}{{series_name}}{% if series_id is defined %} - {{series_id}}{% endif %}{% if trakt_ep_name is defined %} - {{trakt_ep_name}}{% endif %}
              {% elif imdb_name is defined%}{{movie_name}}
              {% else %}{{title}}
              {% endif %}
          via:
            webpost:
              server: <string>
              device_key: <string>
              [level: <string>]
              [badge: <integer>]
              [automatically_copy: <boolean>]
              [copy: <string>]
              [sound: <string>]
              [icon: <string>]
              [group: <string>]
              [is_archive: <boolean>]
              [url: <string>]

    # https://github.com/Finb/bark-server/blob/master/docs/API_V2.md

    """

    schema = {
        'type': 'object',
        'properties': {
            'server': {'type': 'string'},
            'device_key': {'type': 'string'},
            'level': {'type': 'string', 'enum': ['active', 'timeSensitive', 'passive']},
            'badge': {'type': 'integer'},
            'automatically_copy': {'type': 'boolean'},
            'copy': {'type': 'string'},
            'sound': {'type': 'string'},
            'icon': {'type': 'string'},
            'group': {'type': 'string'},
            'is_archive': {'type': 'boolean'},
            'url': {'type': 'string'},
        },
        'required': ['server', 'device_key'],
        'additionalProperties': False,
    }

    def prepare_config(self, config):
        options = config.copy()
        server = options.pop('server')
        if options.pop('automatically_copy', False):
            options['automaticallyCopy'] = '1'
        if options.pop('is_archive', False):
            options['isArchive'] = '1'
        return server, options

    def notify(self, title, message, config):
        """
        Send notification to Bark
        """
        notification = {
            'title': title,
            'body': message,
        }
        server, options = self.prepare_config(config)
        notification.update(options)
        try:
            requests.post(server, json=notification)
        except RequestException as e:
            raise PluginWarning(e.args[0])


@event('plugin.register')
def register_plugin():
    plugin.register(BarkNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



import ssl
from hashlib import sha256

from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.plugin import DependencyError, PluginWarning

plugin_name = 'mqtt'
logger = logger.bind(name=plugin_name)


class MQTTNotifier:
    """
    Example::
      notify:
        entries:
          via:
            - mqtt:
                broker_address: "iot.eclipse.org"
                topic: "flexget/notifications"
                [broker_port: 1883]
                [broker_timeout: 30]
                [broker_transport: ['tcp','websockets'] ]
                [broker_protocol: ['MQTTv31', 'MQTTv311'] ]
                [username: yourUsernameHere]
                [password: yourPasswordHere]
                [encrypted_communication: True/False]
                [certificates:
                    broker_ca_cert: /path/to/pem/encoded/broker_ca_certificate.crt
                    client_cert: /path/to/pem/encoded/client_certificate.crt
                    client_key: /path/to/pem/encoded/client_certificate.key
                    validate_broker_cert: True/False
                    tls_version: ['tlsv1.2', 'tlsv1.1', 'tlsv1']
                ]
                [qos: [0,1,2] ]
                [retain: True/False]
    """

    schema = {
        'type': 'object',
        'properties': {
            'topic': {'type': 'string'},
            'broker_address': {'type': 'string'},
            'broker_port': {'type': 'integer', 'default': 1883},
            'broker_timeout': {'type': 'integer', 'default': 30},
            'broker_transport': {
                'type': 'string',
                'default': 'tcp',
                'enum': ['tcp', 'websockets'],
            },
            'broker_protocol': {
                'type': 'string',
                'default': 'MQTTv311',
                'enum': ['MQTTv31', 'MQTTv311'],
            },
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'encrypted_communication': {'type': 'boolean', 'default': False},
            'certificates': {
                'type': 'object',
                'properties': {
                    'broker_ca_cert': {'type': 'string'},
                    'client_cert': {'type': 'string'},
                    'client_key': {'type': 'string'},
                    'validate_broker_cert': {'type': 'boolean', 'default': True},
                    'tls_version': {'type': 'string', 'enum': ['tlsv1.2', 'tlsv1.1', 'tlsv1', '']},
                },
                'additionalProperties': False,
            },
            'qos': {'type': 'integer', 'minimum': 0, 'maximum': 2, 'default': 0},
            "retain": {"type": "boolean", 'default': False},
        },
        'additionalProperties': False,
        'required': ['broker_address', 'topic'],
        'dependencies': {'password': ['username']},
    }

    def notify(self, title, message, config):
        """
        Publish to an MQTT topic
        """
        try:
            import paho.mqtt.client as mqtt

        except ImportError as e:
            logger.verbose('Error importing paho.mqtt.client: {}', e)
            raise DependencyError(
                plugin_name,
                'paho.mqtt.client',
                f'paho-mqtt python module is required for MQTT notify plugin. ImportError: {e}',
            )

        def on_log_cb(client, userdata, level, buff):
            logger.verbose(str(buff))

        def on_publish_cb(client, userdata, mid):
            logger.verbose(
                'MQTT on_publish callback -  message was successfully published to broker as messageID={}',
                mid,
            )
            client.disconnect()

        def on_disconnect_cb(client, userdata, rc):
            logger.verbose(
                'MQTT on_disconnect callback - disconnected with result code {} [{}]',
                rc,
                conn_rc_description_map.get(rc),
                'Unknown',
            )
            client.loop_stop()

        config['title'] = title
        config['message'] = message
        config['payload'] = f"{config['title']} - {config['message']}"

        conn_rc_description_map = {
            0: 'Connection Accepted',
            1: 'Connection Refused, unacceptable protocol version - The Server does not support the level of the MQTT protocol requested by the Client',
            2: 'Connection Refused, identifier rejected - The Client identifier is correct UTF-8 but not allowed by the Server',
            3: 'Connection Refused, Server unavailable - The Network Connection has been made but the MQTT service is unavailable',
            4: 'Connection Refused, bad user name or password - The data in the user name or password is malformed',
            5: 'Connection Refused, not authorized - The Client is not authorized to connect',
        }

        # Handle the MQTT broker protocol to be used
        if config.get('broker_protocol') == 'MQTTv311':
            config['broker_protocol_class'] = mqtt.MQTTv311
        else:
            config['broker_protocol_class'] = mqtt.MQTTv31

        logger.trace('MQTT notify config={}', config)

        # create the mqtt client
        client = mqtt.Client(
            protocol=config['broker_protocol_class'], transport=config['broker_transport']
        )

        client.enable_logger(logger=logger)

        client.on_log = on_log_cb
        client.on_publish = on_publish_cb
        client.on_disconnect = on_disconnect_cb

        # Handle SSL/TLS communication w/out certificate authentication
        if not config.get('certificates', {}).get('client_cert') and config.get(
            'encrypted_communication'
        ):
            client.tls_set(
                ca_certs=certs.get('broker_ca_cert'),
                certfile=None,
                keyfile=None,
                cert_reqs=ssl.CERT_NONE,
            )

            client.tls_insecure_set(True)

            logger.verbose('Basic SSL/TLS encrypted communications enabled')
            logger.verbose('TLS insecure cert mode enabled. Broker cert will not be validated')

        # Handle SSL/TLS communication with certificate authentication
        if config.get('certificates'):
            certs = config['certificates']
            logger.debug('TLS certificate config: {}', certs)

            tls_version_map = {
                'tlsv1.2': ssl.PROTOCOL_TLSv1_2,
                'tlsv1.1': ssl.PROTOCOL_TLSv1_1,
                'tlsv1': ssl.PROTOCOL_TLSv1,
                '': None,
            }
            tls_version = tls_version_map.get(certs.get('tls_version'), ssl.PROTOCOL_TLSv1_2)
            logger.verbose('TLS version is {}', tls_version)

            cert_required = (
                ssl.CERT_REQUIRED if certs.get('validate_broker_cert', True) else ssl.CERT_NONE
            )

            client.tls_set(
                ca_certs=certs.get('broker_ca_cert'),
                certfile=certs.get('client_cert'),
                keyfile=certs.get('client_key'),
                cert_reqs=cert_required,
                tls_version=tls_version,
            )

            if not certs.get('validate_broker_cert'):
                client.tls_insecure_set(True)
                logger.debug('TLS insecure cert mode enabled. Broker cert will not be validated')
            else:
                logger.debug('TLS secure cert mode enabled. Broker cert will be validated')

        # Handle user/pass authentication
        if config.get('username'):
            logger.debug('Credential passwords s are redacted to protect the innocent...')
            logger.debug(
                'Auth credentials: username=[{}] password sha256 hash is "{}"',
                config.get('username'),
                sha256(str(config.get('password')).encode('utf-8')).hexdigest(),
            )
            logger.debug(
                'You can validate them yourself by calculating the sha256 hex digest of your password string (google is your friend if you do not know how to do this)'
            )
            logger.debug(
                'Note: a password that is not provided (i.e. None) will hash to "{}"',
                sha256(str(None).encode('utf-8')).hexdigest(),
            )

            client.username_pw_set = (config.get('username'), config.get('password'))

        try:
            logger.verbose(
                "Connecting to {}:{}", config.get('broker_address'), config.get('broker_port')
            )
            client.connect(
                config.get('broker_address'),
                config.get('broker_port'),
                config.get('broker_timeout'),
            )
            logger.verbose("Connected to MQTT broker")
        except Exception as e:
            raise PluginWarning(f'Error connecting to MQTT broker: {e}')

        try:
            logger.verbose(
                'Publishing message [{}] to topic [{}] ',
                config.get('payload'),
                config.get('topic'),
            )
            publish_info = client.publish(
                config.get('topic'),
                config.get('payload'),
                qos=config.get('qos'),
                retain=config.get('retain'),
            )
            logger.verbose(
                "Notification sent to broker, waiting for callback response to confirm publishing success - rc={}",
                publish_info,
            )
        except Exception as e:
            raise PluginWarning(f'Error publishing to MQTT broker: {e}')

        client.loop(timeout=config.get('broker_timeout'))
        client.loop_start()


@event('plugin.register')
def register_plugin():
    plugin.register(MQTTNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



import difflib
import json
import random
import re
from datetime import datetime

from loguru import logger

from flexget import plugin
from flexget.utils.requests import Session, TimedLimiter
from flexget.utils.soup import get_soup
from flexget.utils.tools import str_to_int

logger = logger.bind(name='imdb.utils')
# IMDb delivers a version of the page which is unparsable to unknown (and some known) user agents, such as requests'
# Spoof the old urllib user agent to keep results consistent
requests = Session()
requests.headers.update({'User-Agent': 'Python-urllib/2.6'})
# requests.headers.update({'User-Agent': random.choice(USERAGENTS)})

# this makes most of the titles to be returned in english translation, but not all of them
requests.headers.update({'Accept-Language': 'en-US,en;q=0.8'})
requests.headers.update(
    {'X-Forwarded-For': '24.110.%d.%d' % (random.randint(0, 254), random.randint(0, 254))}
)

# give imdb a little break between requests (see: http://flexget.com/ticket/129#comment:1)
requests.add_domain_limiter(TimedLimiter('imdb.com', '3 seconds'))


def is_imdb_url(url):
    """Tests the url to see if it's for imdb.com."""
    if not isinstance(url, str):
        return
    # Probably should use urlparse.
    return re.match(r'https?://[^/]*imdb\.com/', url)


def is_valid_imdb_title_id(value):
    """
    Return True if `value` is a valid IMDB ID for titles (movies, series, etc).
    """
    if not isinstance(value, str):
        raise TypeError("is_valid_imdb_title_id expects a string but got {0}".format(type(value)))
    # IMDB IDs for titles have 'tt' followed by 7 or 8 digits
    return re.match(r'tt\d{7,8}', value) is not None


def is_valid_imdb_person_id(value):
    """
    Return True if `value` is a valid IMDB ID for a person.
    """
    if not isinstance(value, str):
        raise TypeError("is_valid_imdb_person_id expects a string but got {0}".format(type(value)))
    # An IMDB ID for a person is formed by 'nm' followed by 7 digits
    return re.match(r'nm\d{7,8}', value) is not None


def extract_id(url):
    """Return IMDb ID of the given URL. Return None if not valid or if URL is not a string."""
    if not isinstance(url, str):
        return
    m = re.search(r'((?:nm|tt)\d{7,8})', url)
    if m:
        return m.group(1)


def make_url(imdb_id):
    """Return IMDb URL of the given ID"""
    return 'https://www.imdb.com/title/%s/' % imdb_id


class ImdbSearch:
    def __init__(self):
        # de-prioritize aka matches a bit
        self.aka_weight = 0.95
        # prioritize first
        self.first_weight = 1.1
        self.min_match = 0.7
        self.min_diff = 0.01
        self.debug = False

        self.max_results = 50

    def ireplace(self, text, old, new, count=0):
        """Case insensitive string replace"""
        pattern = re.compile(re.escape(old), re.I)
        return re.sub(pattern, new, text, count)

    def smart_match(self, raw_name, single_match=True):
        """Accepts messy name, cleans it and uses information available to make smartest and best match"""
        parser = plugin.get('parsing', 'imdb_search').parse_movie(raw_name)
        name = parser.name
        year = parser.year
        if not name:
            logger.critical('Failed to parse name from {}', raw_name)
            return None
        logger.debug('smart_match name={} year={}', name, str(year))
        return self.best_match(name, year, single_match)

    def best_match(self, name, year=None, single_match=True):
        """Return single movie that best matches name criteria or None"""
        movies = self.search(name)

        if not movies:
            logger.debug('search did not return any movies')
            return None

        # remove all movies below min_match, and different year
        exact = []

        for movie in movies[:]:
            if year and movie.get('year'):
                if movie['year'] != year:
                    logger.debug(
                        'best_match removing {} - {} (wrong year: {})',
                        movie['name'],
                        movie['url'],
                        str(movie['year']),
                    )
                    movies.remove(movie)
                    continue
                # Look for exact match
                if movie['year'] == year and movie['name'].lower() == name.lower():
                    exact.append(movie)
            if movie['match'] < self.min_match:
                logger.debug('best_match removing {} (min_match)', movie['name'])
                movies.remove(movie)
                continue

        if not movies:
            logger.debug('FAILURE: no movies remain')
            return None

        # If we have 1 exact match
        if len(exact) == 1:
            logger.debug('SUCCESS: found exact movie match')
            return exact[0]

        # if only one remains ..
        if len(movies) == 1:
            logger.debug('SUCCESS: only one movie remains')
            return movies[0]

        # check min difference between best two hits
        diff = movies[0]['match'] - movies[1]['match']
        if diff < self.min_diff:
            logger.debug(
                'unable to determine correct movie, min_diff too small (`{}` <-?-> `{}`)',
                movies[0],
                movies[1],
            )
            for m in movies:
                logger.debug('remain: {} (match: {}) {}', m['name'], m['match'], m['url'])
            return None
        else:
            return movies[0] if single_match else movies

    def search(self, name):
        """Return array of movie details (dict)"""
        logger.debug('Searching: {}', name)
        url = 'https://www.imdb.com/find'
        # This may include Shorts and TV series in the results
        params = {'q': name, 's': 'tt'}

        logger.debug('Search query: {}', repr(url))
        page = requests.get(url, params=params)
        actual_url = page.url

        movies = []
        soup = get_soup(page.text)
        # in case we got redirected to movie page (perfect match)
        re_m = re.match(r'.*\.imdb\.com/title/tt\d+/', actual_url)
        if re_m:
            actual_url = re_m.group(0)
            imdb_id = extract_id(actual_url)
            movie_parse = ImdbParser()
            movie_parse.parse(imdb_id, soup=soup)
            logger.debug('Perfect hit. Search got redirected to {}', actual_url)
            movie = {
                'match': 1.0,
                'name': movie_parse.name,
                'imdb_id': imdb_id,
                'url': make_url(imdb_id),
                'year': movie_parse.year,
            }
            movies.append(movie)
            return movies

        section_table = soup.find('table', 'findList')
        if not section_table:
            logger.debug('results table not found')
            return

        rows = section_table.find_all('tr')
        if not rows:
            logger.debug('Titles section does not have links')
        for count, row in enumerate(rows):
            # Title search gives a lot of results, only check the first ones
            if count > self.max_results:
                break

            result_text = row.find('td', 'result_text')
            movie = {}
            additional = re.findall(r'\((.*?)\)', result_text.text)
            if len(additional) > 0:
                if re.match(r'^\d{4}$', additional[-1]):
                    movie['year'] = str_to_int(additional[-1])
                elif len(additional) > 1:
                    movie['year'] = str_to_int(additional[-2])
                    if additional[-1] not in ['TV Movie', 'Video']:
                        logger.debug('skipping {}', result_text.text)
                        continue
            primary_photo = row.find('td', 'primary_photo')
            movie['thumbnail'] = primary_photo.find('a').find('img').get('src')

            link = result_text.find_next('a')
            movie['name'] = link.text
            movie['imdb_id'] = extract_id(link.get('href'))
            movie['url'] = make_url(movie['imdb_id'])
            logger.debug('processing name: {} url: {}', movie['name'], movie['url'])

            # calc & set best matching ratio
            seq = difflib.SequenceMatcher(lambda x: x == ' ', movie['name'].title(), name.title())
            ratio = seq.ratio()

            # check if some of the akas have better ratio
            for aka in link.parent.find_all('i'):
                aka = aka.next.string
                match = re.search(r'".*"', aka)
                if not match:
                    logger.debug('aka `{}` is invalid', aka)
                    continue
                aka = match.group(0).replace('"', '')
                logger.trace('processing aka {}', aka)
                seq = difflib.SequenceMatcher(lambda x: x == ' ', aka.title(), name.title())
                aka_ratio = seq.ratio()
                if aka_ratio > ratio:
                    ratio = aka_ratio * self.aka_weight
                    logger.debug(
                        '- aka `{}` matches better to `{}` ratio {} (weighted to {})',
                        aka,
                        name,
                        aka_ratio,
                        ratio,
                    )

            # prioritize items by position
            position_ratio = (self.first_weight - 1) / (count + 1) + 1
            logger.debug(
                '- prioritizing based on position {} `{}`: {}', count, movie['url'], position_ratio
            )
            ratio *= position_ratio

            # store ratio
            movie['match'] = ratio
            movies.append(movie)

        movies.sort(key=lambda x: x['match'], reverse=True)
        return movies


class ImdbParser:
    """Quick-hack to parse relevant imdb details"""

    def __init__(self):
        self.genres = []
        self.languages = []
        self.actors = {}
        self.directors = {}
        self.writers = {}
        self.score = 0.0
        self.votes = 0
        self.meta_score = 0
        self.year = 0
        self.plot_outline = None
        self.name = None
        self.original_name = None
        self.url = None
        self.imdb_id = None
        self.photo = None
        self.mpaa_rating = ''
        self.plot_keywords = []

    def __str__(self):
        return '<ImdbParser(name=%s,imdb_id=%s)>' % (self.name, self.imdb_id)

    def parse(self, imdb_id, soup=None):
        self.imdb_id = extract_id(imdb_id)
        url = make_url(self.imdb_id)
        self.url = url

        if not soup:
            page = requests.get(url)
            soup = get_soup(page.text)

        data = json.loads(soup.find('script', {'type': 'application/ld+json'}).string)
        if not data:
            raise plugin.PluginError(
                'IMDB parser needs updating, imdb format changed. Please report on Github.'
            )

        props_data = json.loads(soup.find('script', {'type': 'application/json'}).string)
        if (
            not props_data
            or not props_data.get('props')
            or not props_data.get('props').get('pageProps')
        ):
            raise plugin.PluginError(
                'IMDB parser needs updating, imdb props_data format changed. Please report on Github.'
            )

        above_the_fold_data = props_data['props']['pageProps'].get('aboveTheFoldData')
        if not above_the_fold_data:
            raise plugin.PluginError(
                'IMDB parser needs updating, imdb above_the_fold_data format changed. Please report on Github.'
            )

        title = above_the_fold_data.get('titleText')
        if title:
            self.name = title.get('text')
        if not self.name:
            raise plugin.PluginError(
                'IMDB parser needs updating, imdb above_the_fold_data format changed for title. Please report on Github.'
            )

        original_name = above_the_fold_data.get('originalTitleText')
        if original_name:
            self.original_name = original_name.get('text')

        if not self.original_name:
            logger.debug('No original title found for {}', self.imdb_id)

        # NOTE: We cannot use the get default approach here .(get(x, {}))
        # as the data returned in imdb has all fields with null values if they do not exist.
        if above_the_fold_data.get('releaseYear'):
            self.year = above_the_fold_data['releaseYear'].get('year')
        if not self.year:
            logger.debug('No year found for {}', self.imdb_id)

        self.mpaa_rating = data.get('contentRating')
        if not self.mpaa_rating:
            logger.debug('No rating found for {}', self.imdb_id)

        self.photo = data.get('image')
        if not self.photo:
            logger.debug('No photo found for {}', self.imdb_id)

        rating_data = data.get('aggregateRating')
        if rating_data:
            rating_count = rating_data.get('ratingCount')
            if rating_count:
                self.votes = (
                    str_to_int(rating_count) if not isinstance(rating_count, int) else rating_count
                )
            else:
                logger.debug('No votes found for {}', self.imdb_id)

            score = rating_data.get('ratingValue')
            if score:
                self.score = float(score)
            else:
                logger.debug('No score found for {}', self.imdb_id)

        meta_critic = above_the_fold_data.get('metacritic')
        if meta_critic:
            meta_score = meta_critic.get('metascore')
            if meta_score:
                self.meta_score = meta_score.get('score')
        if not self.meta_score:
            logger.debug('No Metacritic score found for {}', self.imdb_id)

        # get director(s)
        directors = data.get('director', [])
        if not isinstance(directors, list):
            directors = [directors]

        for director in directors:
            if director['@type'] != 'Person':
                continue
            director_id = extract_id(director['url'])
            director_name = director['name']
            self.directors[director_id] = director_name

        # get writer(s)
        writers = data.get('creator', [])
        if not isinstance(writers, list):
            writers = [writers]

        for writer in writers:
            if writer['@type'] != 'Person':
                continue
            writer_id = extract_id(writer['url'])
            writer_name = writer['name']
            self.writers[writer_id] = writer_name

        # Details section
        main_column_data = props_data['props']['pageProps'].get('mainColumnData')
        if not main_column_data:
            raise plugin.PluginError(
                'IMDB parser needs updating, imdb main_column_data format changed. Please report on Github.'
            )

        for language in (main_column_data.get('spokenLanguages') or {}).get('spokenLanguages', []):
            self.languages.append(language['text'].lower())

        # Storyline section
        # NOTE: We cannot use the get default approach here .(get(x, {}))
        # as the data returned in imdb has all fields with null values if they do not exist.
        summaries = main_column_data.get('summaries') or {}
        summary_edges = summaries.get('edges') or []
        if len(summary_edges) > 0:
            edge_node = summary_edges[0].get('node') or {}
            plot_text = edge_node.get('plotText') or {}
            # Strip out html
            plot_html = get_soup(plot_text.get('plaidHtml'))
            if plot_html:
                self.plot_outline = plot_html.text
        if not self.plot_outline:
            logger.debug('No storyline found for {}', self.imdb_id)

        storyline_keywords = main_column_data.get('storylineKeywords') or {}
        for keyword_node in storyline_keywords.get('edges') or []:
            keyword = keyword_node.get('node') or {}
            if keyword:
                self.plot_keywords.append(keyword.get('text').lower())

        genres = (above_the_fold_data.get('genres', {}) or {}).get('genres', [])
        self.genres = [g['text'].lower() for g in genres]

        # Cast section
        cast_data = main_column_data.get('cast', {}) or {}
        for cast_node in cast_data.get('edges') or []:
            actor_node = (cast_node.get('node') or {}).get('name') or {}
            actor_id = actor_node.get('id')
            actor_name = (actor_node.get('nameText') or {}).get('text')
            if actor_id and actor_name:
                self.actors[actor_id] = actor_name

        principal_cast_data = main_column_data.get('principalCast', []) or []
        if principal_cast_data:
            for cast_node in principal_cast_data[0].get('credits') or []:
                actor_node = cast_node.get('name') or {}
                actor_id = actor_node.get('id')
                actor_name = (actor_node.get('nameText') or {}).get('text')
                if actor_id and actor_name:
                    self.actors[actor_id] = actor_name



from itertools import zip_longest

from flexget import options
from flexget.event import event
from flexget.terminal import TerminalTable, console
from flexget.utils.qualities import all_components


def do_cli(manager, options):
    components_by_cat = {}
    for component in all_components():
        cat = components_by_cat.setdefault(component.type.title().replace('_', ' '), [])
        cat.append(component)
    for cat_list in components_by_cat.values():
        cat_list.sort(reverse=True)
    header = list(components_by_cat.keys())
    table = TerminalTable(*header, table_type='github')
    for row in zip_longest(*components_by_cat.values(), fillvalue=""):
        table.add_row(*[str(i) for i in row])
    console(table)


@event('options.register')
def register_parser_arguments():
    # Register subcommand
    parser = options.register_command(
        'wiki-qualities',
        do_cli,
        # If we don't specify the help argument, this won't show up in --help, which is good because it's not for users
        # help='Generate the list of qualities for exporting to the wiki.',
    )



from http import HTTPStatus
from urllib.parse import urljoin

from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException

from flexget import plugin
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession

plugin_name = 'ntfysh'

requests = RequestSession(max_retries=3)


class NtfyshNotifier(object):
    """
    Example::

    notify:
      entries:
        via:
          - ntfysh:
              topic: <NTFY_TOPIC>

    Configuration parameters are also supported from entries (eg. through set).
    """

    schema = {
        'type': 'object',
        'properties': {
            'url': {'format': 'url', 'default': 'https://ntfy.sh/'},
            'topic': {'type': 'string'},
            'priority': {'type': 'integer', 'default': 3},
            'delay': {'type': 'string'},
            'tags': {'type': 'string'},
            'username': {'type': 'string'},
            'password': {'type': 'string'},
        },
        'required': ['topic', 'url'],
        'additionalProperties': False,
    }

    def notify(self, title, message, config):
        """
        Send a Ntfy.sh notification
        """
        base_url = config['url']
        topic = config['topic']
        url = urljoin(base_url, topic)

        req = {
            'url': url,
            'data': message,
            'params': {'title': title, 'priority': config['priority']},
        }

        if 'username' in config or 'password' in config:
            req['auth'] = HTTPBasicAuth(config.get('username', ''), config.get('password', ''))

        if 'delay' in config:
            req['params']['delay'] = config['delay']
        if 'tags' in config:
            req['params']['tags'] = config['tags']

        try:
            response = requests.post(**req)
        except RequestException as e:
            if e.response is not None:
                if e.response.status_code in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
                    message = 'Invalid username and password'
                else:
                    message = e.response.text()
            else:
                message = str(e)
            raise PluginWarning(message)


@event('plugin.register')
def register_plugin():
    plugin.register(NtfyshNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.cached_input import cached
from flexget.utils.requests import RequestException

logger = logger.bind(name='my_anime_list')

STATUS = {'watching': 1, 'completed': 2, 'on_hold': 3, 'dropped': 4, 'plan_to_watch': 6, 'all': 7}

AIRING_STATUS = {'airing': 1, 'finished': 2, 'planned': 3, 'all': 6}

ANIME_TYPE = ['all', 'tv', 'ova', 'movie', 'special', 'ona', 'music', 'unknown']


class MyAnimeList:
    """ " Creates entries for series and movies from MyAnimeList list

    Syntax:
    my_anime_list:
      username: <value>
      status:
        - <watching|completed|on_hold|dropped|plan_to_watch>
        - <watching|completed|on_hold|dropped|plan_to_watch>
        ...
      airing_status:
        - <airing|finished|planned>
        - <airing|finished|planned>
        ...
      type:
        - <series|ova...>
    """

    schema = {
        'type': 'object',
        'properties': {
            'username': {'type': 'string'},
            'status': one_or_more(
                {'type': 'string', 'enum': list(STATUS.keys()), 'default': 'all'},
                unique_items=True,
            ),
            'airing_status': one_or_more(
                {'type': 'string', 'enum': list(AIRING_STATUS.keys()), 'default': 'all'},
                unique_items=True,
            ),
            'type': one_or_more(
                {'type': 'string', 'enum': list(ANIME_TYPE), 'default': 'all'}, unique_items=True
            ),
        },
        'required': ['username'],
        'additionalProperties': False,
    }

    @cached('my_anime_list', persist='2 hours')
    def on_task_input(self, task, config):
        selected_status = config.get('status', ['all'])
        if not isinstance(selected_status, list):
            selected_status = [selected_status]

        selected_airing_status = config.get('airing_status', ['all'])
        if not isinstance(selected_airing_status, list):
            selected_airing_status = [selected_airing_status]

        selected_types = config.get('type', ['all'])
        if not isinstance(selected_types, list):
            selected_types = [selected_types]

        selected_status = [STATUS[s] for s in selected_status]
        selected_airing_status = [AIRING_STATUS[s] for s in selected_airing_status]
        list_json = []

        for status in selected_status:
            # JSON pages are limited to 300 entries
            offset = 0
            results = 300
            while results == 300:
                try:
                    list_json += task.requests.get(
                        f'https://myanimelist.net/animelist/{config.get("username")}/load.json',
                        params={'status': status, 'offset': offset},
                    ).json()
                except RequestException as e:
                    logger.error(f'Error finding list on url: {e.request.url}')
                    break
                except (ValueError, TypeError):
                    logger.error('Invalid JSON response')
                    break
                results = len(list_json) or 1
                offset += len(list_json)

            for anime in list_json:
                has_selected_status = (
                    anime["status"] in selected_status or config['status'] == 'all'
                )
                has_selected_airing_status = (
                    anime["anime_airing_status"] in selected_airing_status
                    or config['airing_status'] == 'all'
                )
                has_selected_type = (
                    anime["anime_media_type_string"].lower() in selected_types
                    or config['type'] == 'all'
                )
                if has_selected_status and has_selected_type and has_selected_airing_status:
                    # MAL sometimes returns title as an integer
                    anime['anime_title'] = str(anime['anime_title'])
                    entry = Entry()
                    entry['title'] = anime['anime_title']
                    entry['url'] = f'https://myanimelist.net{anime["anime_url"]}'
                    entry['mal_name'] = anime['anime_title']
                    entry['mal_poster'] = anime['anime_image_path']
                    entry['mal_type'] = anime['anime_media_type_string']
                    entry['mal_tags'] = anime['tags']
                    if entry.isvalid():
                        yield entry


@event('plugin.register')
def register_plugin():
    plugin.register(MyAnimeList, 'my_anime_list', api_ver=2)



import json
import os
import re
import ssl
import xmlrpc.client
from socket import error as socket_error

import requests
from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.template import RenderError

logger = logger.bind(name='aria2')


class RpcClient:
    def __init__(self, server, port, scheme, rpc_path, username, password, secret):
        if secret:
            self.token = 'token:' + secret
        else:
            self.token = ''
        if username and password:
            userpass = '%s:%s@' % (username, password)
        else:
            userpass = ''
        self.url = '%s://%s%s:%s/%s' % (scheme, userpass, server, port, rpc_path)
        logger.debug('aria2 url: {}', self.url)

    def add_uri(self, uris, options):
        raise plugin.PluginError('Unsupported Operation')

    def add_torrent(self, torrent, options):
        raise plugin.PluginError('Unsupported Operation')


class JsonRpcClient(RpcClient):
    RPC_ID = 'FLEXGET'
    RPC_VERSION = '2.0'

    ADDURI_METHOD = 'aria2.addUri'
    GETGLOBALSTAT_METHOD = 'aria2.getGlobalStat'
    ADDTORRENT_METHOD = 'aria2.addTorrent'
    ADDMETALINK_METHOD = 'aria2.addMetalink'

    def __init__(self, server, port, scheme, rpc_path, username=None, password=None, secret=None):
        super(JsonRpcClient, self).__init__(
            server, port, scheme, rpc_path, username, password, secret
        )
        # trigger _default_error_handle on failure
        self.get_global_stat()

    def _get_req_params(self, method, params=[]):
        req_params = {
            'jsonrpc': JsonRpcClient.RPC_VERSION,
            'id': JsonRpcClient.RPC_ID,
            'method': method,
            'params': params,
        }
        if self.token:
            req_params['params'].insert(0, self.token)
        if not req_params['params']:
            del req_params['params']
        return req_params

    def _default_error_handle(code, message):
        logger.critical('Fault code {} message {}', code, message)
        raise plugin.PluginError('Fault code %s message %s' % (code, message), logger)

    def _default_success_handle(response):
        return response.text

    def _post(
        self, method, params, on_success=_default_success_handle, on_fail=_default_error_handle
    ):
        resp = requests.post(self.url, data=json.dumps(self._get_req_params(method, params)))
        result = resp.json()
        if "error" in result:
            return on_fail(result["error"]["code"], result["error"]["message"])
        else:
            return on_success(resp)

    def add_uri(self, uris, options):
        # https://aria2.github.io/manual/en/html/aria2c.html#aria2.addUri
        return self._post(JsonRpcClient.ADDURI_METHOD, params=[[uris], options])

    def add_torrent(self, torrent, options):
        # https://aria2.github.io/manual/en/html/aria2c.html#aria2.addTorrent
        params = [torrent]
        return self._post(JsonRpcClient.ADDTORRENT_METHOD, params=[torrent, options])

    def get_global_stat(self):
        # https://aria2.github.io/manual/en/html/aria2c.html#aria2.getGlobalStat
        return self._post(JsonRpcClient.GETGLOBALSTAT_METHOD, params=[])


class XmlRpcClient(RpcClient):
    def __init__(self, server, port, scheme, rpc_path, username=None, password=None, secret=None):
        schemes = {'http': None, 'https': ssl.SSLContext()}
        if scheme not in schemes:
            raise plugin.PluginError('Unknown scheme: %s' % (scheme), logger)
        super(XmlRpcClient, self).__init__(
            server, port, scheme, rpc_path, username, password, secret
        )
        try:
            self._aria2 = xmlrpc.client.ServerProxy(self.url, context=schemes[scheme]).aria2
        except xmlrpc.client.ProtocolError as err:
            raise plugin.PluginError(
                'Could not connect to aria2 at %s. Protocol error %s: %s'
                % (self.url, err.errcode, err.errmsg),
                logger,
            )
        except xmlrpc.client.Fault as err:
            raise plugin.PluginError(
                'XML-RPC fault: Unable to connect to aria2 daemon at %s: %s'
                % (self.url, err.faultString),
                logger,
            )
        except socket_error as e:
            raise plugin.PluginError(
                'Socket connection issue with aria2 daemon at %s: %s' % (self.url, e), logger
            )
        except:
            logger.opt(exception=True).debug('Unexpected error during aria2 connection')
            raise plugin.PluginError(
                'Unidentified error during connection to aria2 daemon', logger
            )

    def add_uri(self, uris, options):
        # https://aria2.github.io/manual/en/html/aria2c.html#aria2.addUri
        params = [[uris]]
        if options:
            params.append(options)
        if self.token:
            params.insert(0, self.token)
        return self._aria2.addUri(*params)

    def add_torrent(self, torrent, options):
        # https://aria2.github.io/manual/en/html/aria2c.html#aria2.addTorrent
        params = [torrent]
        if options:
            params.append(options)
        if self.token:
            params.insert(0, self.token)
        return self._aria2.addTorrent(*params)


class OutputAria2:
    """
    Simple Aria2 output

    Example::

        aria2:
          path: ~/downloads/

    """

    schema = {
        'type': 'object',
        'properties': {
            'server': {'type': 'string', 'default': 'localhost'},
            'port': {'type': 'integer', 'default': 6800},
            'secret': {'type': 'string', 'default': ''},
            # NOTE: To be deprecated by aria2
            'username': {'type': 'string', 'default': ''},
            'password': {'type': 'string', 'default': ''},
            'scheme': {'type': 'string', 'default': 'http'},
            # NOTE: xml/json
            'rpc_mode': {'type': 'string', 'default': 'xml'},
            'rpc_path': {'type': 'string', 'default': 'rpc'},
            'path': {'type': 'string'},
            'filename': {'type': 'string'},
            'add_extension': {
                'oneOf': [
                    {'type': 'string'},
                    {'type': 'boolean'},
                ],
                'default': 'no',
            },
            'options': {
                'type': 'object',
                'additionalProperties': {
                    'oneOf': [{'type': 'string'}, {'type': 'number'}, {'type': 'boolean'}]
                },
            },
        },
        'required': ['path'],
        'additionalProperties': False,
    }

    def prepare_config(self, config):
        config.setdefault('server', 'localhost')
        config.setdefault('port', 6800)
        config.setdefault('username', '')
        config.setdefault('password', '')
        config.setdefault('scheme', 'http')
        config.setdefault('rpc_mode', 'xml')
        config.setdefault('rpc_path', 'rpc')
        config.setdefault('secret', '')
        config.setdefault('options', {})
        config.setdefault('add_extension', False)
        options = config['options']
        for key in options:
            if isinstance(options[key], bool):
                options[key] = str(options[key]).lower()
            elif not isinstance(options[key], str):
                options[key] = str(options[key])
        return config

    def on_task_output(self, task, config):
        # don't add when learning
        if task.options.learn:
            return
        config = self.prepare_config(config)
        rpc_clients = {'xml': XmlRpcClient, 'json': JsonRpcClient}
        if config['rpc_mode'] not in rpc_clients:
            entry.fail('Unknown rpc_mode: %s' % config['rpc_mode'])
        aria2 = rpc_clients[config['rpc_mode']](
            config['server'],
            config['port'],
            config['scheme'],
            config['rpc_path'],
            config['username'],
            config['password'],
            config['secret'],
        )

        for entry in task.accepted:
            if task.options.test:
                logger.verbose('Would add `{}` to aria2.', entry['title'])
                continue
            try:
                self.add_entry(aria2, entry, config, task)
            except socket_error as se:
                entry.fail('Unable to reach Aria2: %s' % se)
            except xmlrpc.client.Fault as err:
                logger.critical('Fault code {} message {}', err.faultCode, err.faultString)
                entry.fail('Aria2 communication Fault')
            except Exception as e:
                logger.opt(exception=True).debug('Exception type {}', type(e))
                raise

    def add_entry(self, aria2: RpcClient, entry, config, task):
        """
        Add entry to Aria2
        """
        options = config['options']
        try:
            path = entry.get('path', config.get('path', None))
            options['dir'] = os.path.expanduser(entry.render(path).rstrip('/'))
        except RenderError as e:
            entry.fail('failed to render \'path\': %s' % e)
            return

        filename = entry.get('content_filename', config.get('filename', None))
        add_extension = entry.get('content_extension', config.get('add_extension', False))

        if filename:
            if add_extension:
                ext = None
                if isinstance(add_extension, bool):
                    logger.debug('Getting filename from `{}`', entry['url'])
                    content_disposition = None
                    try:
                        with task.requests.get(
                            entry['url'], headers=None, stream=True
                        ) as response:
                            content_disposition = response.headers.get('content-disposition', None)
                    except Exception as e:
                        logger.warning('Not possible to retrive file info from `{}`', entry['url'])
                        entry.fail('Not possible to retrive file info from `%s`' % entry['url'])
                        return

                    if content_disposition:
                        fname_match = re.findall(
                            r'filename=["\']?([^"\']+)["\']?', content_disposition
                        )
                        if fname_match:
                            fname = fname_match[0]
                            fname_info = os.path.splitext(fname)
                            if len(fname_info) == 2:
                                ext = fname_info[1]
                                logger.debug(
                                    'Filename from `{}` is {} with ext `{}`',
                                    entry['url'],
                                    fname,
                                    ext,
                                )

                else:
                    ext = add_extension if add_extension[0] == '.' else '.' + add_extension

                if not ext:
                    logger.warning('Not possible to retrive extension')
                    entry.fail('Not possible to retrive extension')
                    return

                logger.debug('Adding extension `{}` to file `{}`', ext, filename)

                filename += ext

            try:
                options['out'] = os.path.expanduser(entry.render(filename))
            except RenderError as e:
                entry.fail('failed to render \'filename\': %s' % e)
                return

        # handle torrent files
        if 'torrent' in entry:
            if 'file' in entry:
                torrent_file = entry['file']
            elif 'location' in entry:
                # in case download plugin moved the file elsewhere
                torrent_file = entry['location']
            else:
                entry.fail('Cannot find torrent file')
                return
            return aria2.add_torrent(
                xmlrpc.client.Binary(open(torrent_file, mode='rb').read()), options
            )
        # handle everything else (except metalink -- which is unsupported)
        # so magnets, https, http, ftp .. etc
        return aria2.add_uri(entry['url'], options)


@event('plugin.register')
def register_plugin():
    plugin.register(OutputAria2, 'aria2', api_ver=2)



import re

from loguru import logger

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.components.sites.utils import torrent_availability
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.requests import TokenBucketLimiter
from flexget.utils.soup import get_soup

logger = logger.bind(name='nCore')

URL = 'https://ncore.pro'

HEADERS = {'Content-type': 'application/x-www-form-urlencoded'}

PASSKEY = ''

CATEGORIES = {
    "all",
    "xvid_hun",
    "xvid",
    "dvd_hun",
    "dvd",
    "dvd9_hun",
    "dvd9",
    "hd_hun",
    "hd",
    "xvidser_hun",
    "xvidser",
    "dvdser_hun",
    "dvdser",
    "hdser_hun",
    "hdser",
}

SORT = {
    'default': 'fid',
    'date': 'fid',
    'size': 'size',
    'seeds': 'seeders',
    'leechers': 'leechers',
    'downloads': 'times_completed',
}


class UrlRewriteNcore:
    """nCore urlrewriter."""

    schema = {
        'type': 'object',
        'properties': {
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'category': {
                'type': 'array',
                'items': {'type': 'string', 'enum': list(CATEGORIES)},
                'default': [],
            },
            'sort_by': {'type': 'string', 'default': 'default', 'enum': list(SORT)},
            'sort_reverse': {'type': 'boolean', 'default': False},
        },
        'required': ['username', 'password'],
        'additionalProperties': False,
    }

    request_limiter = TokenBucketLimiter('ncore.pro/', 100, '5 seconds')

    @plugin.internet(logger)
    def search(self, task, entry, config):
        """
        Search for name from ncore.
        """

        data = {
            "set_lang": "hu",
            "submitted": "1",
            "nev": config.get("username"),
            "pass": config.get("password"),
            "ne_leptessen_ki": 0,
        }

        page = task.requests.post(URL + "/login.php", data=data, headers=HEADERS)
        soup = get_soup(page.content)
        passkey_line = str(soup.find('link', href=re.compile(r'rss\.php\?key=')))
        PASSKEY = passkey_line[passkey_line.find("key=") : passkey_line.find('"', 20, 90)]

        entries = set()

        for search_string in entry.get('search_strings', [entry['title']]):
            data = {
                "mire": search_string,
                "miben": "name",
                "miszerint": SORT[config.get('sort_by')],
            }

            if config.get('category'):
                data["kivalasztott_tipus[]"] = config.get('category')
                data["tipus"] = "kivalasztottak_kozott"
            else:
                data["tipus"] = "all_own"

            if config.get('sort_reverse'):
                data["hogyan"] = "DESC"

            page = task.requests.post(URL + "/torrents.php", data=data, headers=HEADERS)
            soup = get_soup(page.content)
            for a in soup.findAll('a', title=re.compile(".+")):
                if 'details' in a.get('href'):
                    e = Entry()

                    href = a.get('href')
                    id = href[href.find("id=") + 3 :]

                    e['title'] = a.text
                    e['url'] = URL + '/torrents.php?action=download&id={0}&'.format(id) + PASSKEY

                    parent = a.parent.parent.parent.parent

                    e['torrent_seeds'] = parent.find('div', class_='box_s2').string
                    e['torrent_leeches'] = parent.find('div', class_='box_l2').string
                    e['content_size'] = parent.find('div', class_='box_meret2').string
                    e['torrent_availability'] = torrent_availability(
                        e['torrent_seeds'], e['torrent_leeches']
                    )

                    yield e


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteNcore, 'ncore', interfaces=['search'], api_ver=2)



from enum import Enum, unique

from loguru import logger

from flexget import db_schema, plugin
from flexget.components.sites.utils import torrent_availability
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.requests import RequestException
from flexget.utils.requests import Session as RequestSession
from flexget.utils.soup import get_soup

logger = logger.bind(name='hebits')
Base = db_schema.versioned_base('hebits', 0)

requests = RequestSession()


@unique
class HeBitsCategory(Enum):
    movies = 1
    tv = 2
    theater = 3
    software = 4
    games = 5
    music = 6
    books = 7
    movies_packs = 8
    porno = 9
    other = 10


class HeBitsSort(Enum):
    time = 1
    year = 2
    size = 3
    snatched = 4
    seeders = 5
    leechers = 6
    random = 7


class SearchHeBits:
    """
    HeBits search plugin.
    """

    schema = {
        'type': 'object',
        'properties': {
            'userid': {'type': 'number'},
            'session': {'type': 'string'},
            'category': {'type': 'string', 'enum': [value.name for value in HeBitsCategory]},
            'free': {'type': 'boolean'},
            'double': {'type': 'boolean'},
            'triple': {'type': 'boolean'},
            'order_by': {
                'type': 'string',
                'enum': [value.name for value in HeBitsSort],
                'default': HeBitsSort.time.name,
            },
            'order_desc': {'type': 'boolean', 'default': True},
        },
        'required': ['userid', 'session'],
        'additionalProperties': False,
    }

    base_url = 'https://hebits.net/'
    search_url = f'{base_url}ajax.php'
    download_url = f'{base_url}torrents.php'
    user_profile_url = f'{base_url}user.php'

    @staticmethod
    def _fetch_account_info(url, config) -> dict:
        logger.debug('Trying to fetch hebits passkey and authkey from user profile')
        authkey, passkey = None, None
        cookies = {"userid": f"{config['userid']}", "session": f"{config['session']}"}
        response = requests.get(url, cookies=cookies, params={'id': config['userid']})
        user_profile_soup = get_soup(response.text)
        for tag in user_profile_soup.find_all("meta"):
            if tag.get("name", None) == "authkey":
                authkey = tag.get('content')
        for tag in user_profile_soup.find_all("a"):
            if tag.get("id", None) == "passkey":
                passkey = tag.get('onclick').split('\'')[1]
        if not authkey or not passkey:
            raise plugin.PluginError('Could not fetch authkey or passkey, layout change?')
        return {
            'authkey': authkey,
            'passkey': passkey,
        }

    @plugin.internet(logger)
    def search(self, task, entry, config):
        """Search for entries on HeBits"""

        account_info = self._fetch_account_info(self.user_profile_url, config)
        params = {
            'action': 'browse',
            'group_results': 0,
        }
        cookies = {"userid": f"{config['userid']}", "session": f"{config['session']}"}

        if 'category' in config:
            cat = HeBitsCategory[config['category']].value
            params[f'filter_cat[{cat}]'] = 1

        entries = set()
        params['order_by'] = HeBitsSort[config['order_by']].value
        params['order_way'] = 'desc' if config['order_desc'] else 'asc'

        for search_string in entry.get(
            'search_strings', [entry['title'], entry.get('original_title'), entry.get('imdb_id')]
        ):
            params['searchstr'] = search_string
            logger.debug('Using search params: {}', params)
            try:
                page = requests.get(self.search_url, cookies=cookies, params=params)
                page.raise_for_status()
                search_results_response = page.json()
            except RequestException as e:
                logger.error('HeBits request failed: {}', e)
                continue
            if search_results_response['status'] != 'success':
                logger.error('HeBits request failed: server error')
                continue

            search_results = search_results_response['response']['results']
            for result in search_results:
                torrent_id = result['torrents'][0]['torrentId']
                seeders = result['torrents'][0]['seeders']
                leechers = result['torrents'][0]['leechers']
                size = result['torrents'][0]['size'] / 2**20
                title = result['torrents'][0]['release']

                entry = Entry(
                    torrent_seeds=seeders,
                    torrent_leeches=leechers,
                    torrent_availability=torrent_availability(seeders, leechers),
                    content_size=size,
                    title=title,
                    torrent_freeleech=result['torrents'][0]['isFreeleech'],
                    torrent_triple_up=result['torrents'][0]['isUploadX3'],
                    torrent_double_up=result['torrents'][0]['isUploadX2'],
                    url=f"{self.download_url}?action=download&id={torrent_id}&authkey={account_info['authkey']}&torrent_pass={account_info['passkey']}",
                )
                entries.add(entry)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(SearchHeBits, 'hebits', interfaces=['search'], api_ver=2)
